########################################################################
#
# Menu for all elements
# (C) 1997 Harald Kirsch (kir@iitb.fhg.de)
#
# $Revision: 2.6 $, $Date: 1999/12/16 22:14:06 $
#
########################################################################

########################################################################
# 
# elePrimary -- make an element the primary element
#
# Parameter which must be the index within $w.eleMenu.m to which all
# this refers.
#
proc ::nep::elePrimary {w which} {
  variable current
  variable fdof
  variable bsof
  variable cbi
  variable dx

  ## reconfigure name of element menu
  set label [$w.eleMenu.m entrycget $which -label]
  set color [$w.eleMenu.m entrycget $which -foreground]
  $w.eleMenu configure \
      -text $label \
      -foreground $color \
      -activeforeground $color

  ## extract the element bound to `which'
  set element [$w.eleMenu.m entrycget $which -command]


  ## Here we make it the primary element. We set it in current and
  ## change the block-relative distance dx.
  set current($w) $element
  calculateCbiAndDx $w

  #$w.g element deactivate *
  catch "destroy $w.g.pos"

  ## set axis labels
  set fd $fdof($w,$element)
  $w.g xaxis configure -title [::tdfio::get $fd Xlabel {samples [-]}]
  $w.g yaxis configure -title [::tdfio::get $fd Ylabel {value [-]}]
  $w.g configure -title [::tdfio::get $fd Comment {a plot}]

  ## rearrange display list
  set l [$w.g element show]
  set epos  [lsearch $l $element]
  set l [lreplace $l $epos $epos]
  lappend l $element
  $w.g element show $l

}
########################################################################
#
# This is the postcommand for $w.eleMenu.m.m. It configures the
# elements of this submenu to relate it to the cascade button which
# invoked it.
#
proc ::nep::setupEleSubmenu {w} {
  set which [$w.eleMenu.m index active]

  set color [$w.eleMenu.m entrycget $which -foreground]
  set label [$w.eleMenu.m entrycget $which -label]
  $w.eleMenu.m.m configure \
      -disabledforeground $color 

  $w.eleMenu.m.m entryconfigure 0 \
      -label $label \
      -state disabled

  set mm $w.eleMenu.m.m
  $mm entryconfigure Primary -command "::nep::elePrimary {$w} $which"
  $mm entryconfigure Delete -command "::nep::deletePlot {$w} $which"
  $mm entryconfigure Reload* -command "::nep::reloadFile {$w} $which"
  $mm entryconfigure "Configure ..." \
      -command "::nep::configPrimary {$w} $which"
}
########################################################################
proc ::nep::eleMenuCreate {w} {
  menu $w.eleMenu.m \
      -tearoff 0

  ## Now create the menu which is affixed to every entry which is
  ## dynamically added as a cascade button to the above menu
  set mm $w.eleMenu.m.m
  menu $mm \
      -tearoff 0 \
      -postcommand "::nep::setupEleSubmenu {$w}"

  
  ## Here we just prepare the look of the menu. Commands are added on
  ## the fly with the postcommand above.
  $mm add command -label x
  $mm add command -label Primary 
  $mm add command -label Delete
  $mm add command -label "Reload (r)"
  $mm add command -label "Configure ..." 
}
########################################################################
#
# Add an entry to the element menu
#
proc ::nep::eleMenuAdd {w element text} {
  set elemColor [$w.g element cget $element -color] 

  ## The -command option does not really denote a command. It is
  ## rather used like a widget variable.
  $w.eleMenu.m add cascade \
      -label $text \
      -command $element \
      -foreground $elemColor \
      -activeforeground $elemColor \
      -menu $w.eleMenu.m.m
  elePrimary $w end
}
########################################################################
