########################################################################
# 
# handles additional items to be placed around a plot
#
# $Revision: 2.8 $, $Date: 1999/12/16 22:14:07 $
#
########################################################################

namespace eval ::nep::items {
  ## An array which contains for each of nep's base windows $w a map
  ## from file identifiers (inode/dev-combinations) called fid below
  ## to unique marker ids.
  variable idof

  ## An array which contains for each of nep's base windows $w and
  ## each $fid an entry used for checkbuttons.
  variable visible

  ## An array which contains for each of nep's base windows $w and
  ## each $fid the name of the file associated with fid
  variable fileof
}

########################################################################
#
# This is a function intended to be called from within an item file
#
proc nepRangeItem {from to {name {}} } {
  upvar w w  color color  ID ID
  
  set coords [list $from -Inf $to -Inf]
  #puts $coords

  $w.g marker create line \
      -name item.$ID.[::nep::uniqueID] \
      -outline $color \
      -linewidth 1 \
      -yoffset -10 \
      -coords $coords 

  if {"$name"==""} return
  set x [expr 0.5*($from+$to)]
  $w.g marker create text \
      -text $name \
      -name item.$ID.[::nep::uniqueID] \
      -foreground $color \
      -background black \
      -yoffset -10 \
      -pady "0 0" \
      -anchor s \
      -coords "$x -Inf"
}
########################################################################
proc ::nep::items::jumpto {w lbox y} {
  set which [$lbox nearest $y]
  set entry [$lbox get $which]

  #puts $entry
  regexp {\(([^ ]+) ([^ ]+)\)} $entry dummy x y
  #puts "$x $y"

  ## We align the middle of the x-axis to the middle of the item
  foreach {xmin xmax} [$w.g axis limits x] break
  set dx [expr {$xmax-$xmin}]
  set min [expr $x-0.5*$dx]
  set max [expr $min+$dx]
  $w.g xaxis configure -min $min -max $max
  ::nep::calculateCbiAndDx $w
  #setCurblk $w [currentXvector $w] $min

}
  
########################################################################
##
## Pops up a text-widget which lists all items for fast reference
##
proc ::nep::items::listem {w fid} {
  variable idof

  if {![winfo exists $w.itemsList]} {
    toplevel $w.itemsList
    listbox $w.itemsList.l \
	-yscrollcommand "$w.itemsList.sb set"
    scrollbar $w.itemsList.sb \
	-command "$w.itemsList.l yview"

    button $w.itemsList.done \
	-text Done \
	-command "wm withdraw $w.itemsList" 
    grid $w.itemsList.l -row 0 -column 0 \
	-sticky nsew
    grid $w.itemsList.sb -row 0 -column 1 \
	-sticky ns
    grid $w.itemsList.done -row 1 -column 0 -columnspan 2
    grid columnconfigure $w.itemsList 0 -weight 1
    grid rowconfigure $w.itemsList 0 -weight 1
    bind $w.itemsList.l <1> "::nep::items::jumpto {$w} %W %y"
    balloonHelp $w.itemsList.l "Button-1 jumps to item"
  } else {
    wm deiconify $w.itemsList
  }
  
  #array set a $nepP($w,item,$fname)
  $w.itemsList.l delete 0 end
  set ID $idof($w,$fid)
  foreach item  [$w.g marker names item.$ID.*] {
    if {"text"!="[$w.g marker type $item]"} continue
    set text [$w.g marker cget $item -text]
    set coords [$w.g marker cget $item -coords]
    $w.itemsList.l insert end "\"$text\" @ ($coords)"
  }
}
########################################################################
##
## This functions is called just once from the main of nep to create a 
## basic and empty menu for items.
##
proc ::nep::items::createMenu {w m} {
  ## Create the menu with two initial entries to switch all items on
  ## or off.
  menu $m -tearoff 0 
  $m add command -label "all visible" \
      -command [list ::nep::items::allVisible $w 1] 
  $m add command -label "all invisible" \
      -command [list ::nep::items::allVisible $w 0] 

  ## Every cascade element for items which will be dynamically added
  ## to the above menu will post the following menu. However, right
  ## before the menu is posted, it is configured to reflect the
  ## cascade of the above menu to which it refers.
  set mm $m.m
  menu $mm -tearoff 0 \
      -postcommand [list ::nep::items::postSubmenu $w $m]

  $mm add command               ;#will be just the title
  $mm add checkbutton \
      -label visible \
      -variable ::nep::items::dummy($w)
  $mm add command -label List
  $mm add command -label Delete
  $mm add command -label Reload
}
########################################################################
##
##
proc ::nep::items::postSubmenu {w m} {
  variable fileof

  set which [$m index active]

  ## Give the submenu the right look by adapting the topmost element
  ## used as a label.
  set color [$w.items.m entrycget $which -foreground]
  set label [$w.items.m entrycget $which -label]

  $m.m configure -disabledforeground $color 
  $m.m entryconfigure 0 \
      -label $label \
      -state disabled


  ## get the file identifier associated with this menu entry
  set fid [$w.items.m entrycget $which -command]

  $m.m entryconfigure visible \
      -variable ::nep::items::visible($w,$fid) \
      -command [list ::nep::items::visible $w $fid]
  $m.m entryconfigure List \
      -command [list ::nep::items::listem $w $fid]
  $m.m entryconfigure Delete \
      -command [list ::nep::items::delete $w $fid]
  $m.m entryconfigure Reload \
      -command [list ::nep::items::add $w $fileof($w,$fid)]
}
########################################################################
proc ::nep::items::delete {w fid} {
  variable visible
  variable idof
  variable fileof

  ## Delete all markers from the graph
  set ID $idof($w,$fid)
  set l [$w.g marker names item.$ID.*]
  eval $w.g marker delete $l

  ## Delete the item's entry in the menu. This is a bit tricky because
  ## we have to perfrom kind of a reverse-lookup
  set L [$w.items.m index end]
  for {set i 2} {$i<=$L} {incr i} {
    set f [$w.items.m entrycget $i -command]
    if {"$f"=="$fid"} {
      set color [$w.items.m entrycget $i -foreground]
      incr ::nep::colors($color) -1
      $w.items.m delete $i
      break
    }
  }
  if {"[$w.items.m index end]"=="1"} {
    $w.items configure -state disabled
  }

  ## Throw away the associated variables. In particular the variable
  ## for the menus checkbutton is a bit sticky, if we don't
  ## reconfigure the checkbutton
  $w.items.m.m entryconfigure visible -variable ::nep::items::dummy($w)
  unset visible($w,$fid)
  unset idof($w,$fid)
  unset fileof($w,$fid)
}
########################################################################
proc ::nep::items::visible {w fid} {
  variable idof
  variable visible

  set ID $idof($w,$fid)
  foreach item [$w.g marker names item.$ID.*] {
    $w.g marker configure $item -hide [expr {!$visible($w,$fid)}]
  }
}
########################################################################
proc ::nep::items::allVisible {w state} {
  variable idof
  variable visible

  foreach x [array names idof $w,*] {
    set l [split $x ,]
    set fid [lindex $l 1]
    set visible($w,$fid) $state
    visible $w $fid
  }
}
########################################################################

proc ::nep::items::add {w fname} {
  variable idof
  variable visible
  variable fileof

  ## Try a guess about whether we know already about this file
  file stat $fname info
  set fid "$info(ino)-$info(dev)"
  catch {parray idof}
  if {[info exist idof($w,$fid)]} {
    delete $w $fid
  }

  set fileof($w,$fid) $fname
  set idof($w,$fid) [::nep::uniqueID]	;# ID used for markers on the plot
  set visible($w,$fid) 1

  set ID $idof($w,$fid)
  set label [file tail $fname]
  set color [::nep::findColor]

  ## This is dangerous, but I expect someone using nep with items to
  ## have some basic knowledge about tcl.
  source $fname

  ## Prepare the menu entry. The -command option is not really
  ## used as a command, but as a entry-specific variable.
  $w.items.m add cascade \
      -menu $w.items.m.m \
      -label $label \
      -foreground $color \
      -activeforeground $color \
      -command $fid
  set nepP($w,item,$fname,b) 1
  $w.items configure -state normal

  #lappend nepP($w,item,$fname) ID $ID  color $color
}
########################################################################
