########################################################################
#
# nep -- Noch Ein Plotprogramm
#
# The whole application can be called as a proc, if this file is
# sourced in. Care has been take to allow even multiple calls of `nep'
# to be able to run multiple independent nep-windows from one
# application. (However, this is not yet tested.)
#
# USAGE:
#  nep w files geom
#        w -- name of nep-window to create
#    files -- list of files to read
#     geom -- geometry
# 
# $Revision: 2.20 $, $Date: 2000/01/17 16:26:41 $
#
########################################################################

## make sure, that this is not accidently sourced in twice
if {[llength [info proc nep]]} {return 0}

## At this point, some kind soul must have set the variable
## ::nep::base already, otherwise we are out of luck
foreach file {tdfio eleMenu configure items} {
  source [file join $::nep::base $file.tcl]
}
foreach file {welDa balloonHelp balloonMsg} {
  source [file join $::nep::base tclutil $file.tcl]
}
if {[file readable [file join $::nep::base tclutil teacher.tcl]]} {
  source [file join $::nep::base tclutil teacher.tcl]
}

## My private party
option add *g.plotBackground black widgetDefault
option add *g.plotBorderWidth 0 widgetDefault
option add *g.Legend.relief flat widgetDefault
option add *g.Legend.Background lightgrey widgetDefault
option add *g.TextMarker.Foreground white widgetDefault
option add *xsel.highlightColor      red widgetDefault
option add *ysel.highlightColor      red widgetDefault
option add *Graph.Cursor \
    [list @$::nep::base/moveCursor.bm \
	 $::nep::base/moveCursorM.bm yellow black] \
    widgetDefault



namespace eval ::nep {

  # namespace variables

  # The path were to find bitmaps and such
  variable base

  # In the following, $w indicates a nep widget. 

  # an array which contains for every $w the graph element
  # which is the primary element, i.e. on which most operations are
  # applied. 
  variable current

  # an array which contains for every $w and every graph element the
  # natural blocksize used for paging. It is initialized from the data
  # file when it is opened.
  variable bsof

  # an array which contains for every $w an index into the primary
  # signal which is taken as the beginning of a block of this
  # signal. The following constraints are maintained for any cbi:
  # -> cbi%bsof(primary plot)==0
  # -> cbi>=0
  # -> cbi<count(primary plot)
  variable cbi;				# current block index

  # an array which contains for every $w the distance on the x-axis
  # between the beginning of a block (cbi) within the signal and
  # minimum of the x-axis. Many operations for moving the viewport
  # maintain this distance as well as the length of the x-axis.
  variable dx

  # an array which contains for every $w a map from graph elements to
  # file handles of the files they are representing.
  variable fdof

  # an array which contains for every $w a map from graph elements to
  # IDs of item files as returned by ::nep::items::Add
  variable itmof

  # last unique number used.
  variable nextID 0

  # an array which contains for every $w a flag which tells us, if an
  # update of the x-axis is under way. This allows to throw repeated
  # keypresses away
  variable repeat

  # an array which contains for every $w the stack of zoomed states
  variable zstack

  # an array which contains for every $w an element to control modal
  # dialogs 
  variable modalWait

  #
  # colors
  # -- to use for plots. In fact this array contains for the possible
  # colors a use count. Whenever a new color is needed, the one with
  # the smallest use count is taken.
  variable colors
  set colors(all) { 
    \#ff3333 \#33ff33 \#3333ff
    \#ffff33 \#ff33ff \#33ffff
    \#ff9933 \#ff3399
    \#3399ff \#9933ff
    \#33ff99 \#99ff33
    \#ff8800 \#aaff00 \#00ffaa
    \#ff00aa \#8800ff \#3388ff
    \#AAAA00 \#AA00AA \#00AAAA
  } 
  foreach c $colors(all) {
    ## mark color as unused.
    set colors($c) 0
  }
  unset c
}
########################################################################
#
# returns the color with the smallest use count
#
proc ::nep::findColor {} {
  variable colors

  set min 111111
  foreach c $colors(all) {
    if {$colors($c)<$min} {
      set res $c
      set min $colors($c)
    }
  }
  incr colors($res) 
  return $res
}
########################################################################
#
proc ::nep::uniqueID {} {
  variable nextID
  return [incr nextID]
}
########################################################################
#
# return the vector for the given axis of the primary plot element
#
# axis must be either x or y
#
proc ::nep::currentVec {w axis} {
  variable current
  return [$w.g element cget $current($w) -${axis}data]
}
########################################################################
##
## The x-axis was just changed so quite a few things must be changed.
##
proc ::nep::updateDisplay {w} {
  variable fdof

  foreach {xmin xmax} [$w.g axis limits x] break

  # parray nepP
  foreach ele [$w.g element names] {
    set id [string range $ele 1 end]
    ::tdfio::read $fdof($w,$ele) $xmin $xmax ::nep::X$id ::nep::Y$id
  }
  update
}
########################################################################
#
# Never ever change cbi except through this function
#
proc ::nep::setCbi {w val} {
  variable cbi
  variable current
  variable bsof

  set cbi($w) $val

  ## adjust curblk
  set blk [expr {$val/$bsof($w,$current($w))}]
  $w.curblk delete 0 end
  $w.curblk insert end $blk
}
########################################################################
#
# one of cbi or dx was changed. Here we adjust xmin and xmax while
# maintaining xmax-xmin. Then updateDisplay is called to read the
# necessary data in. In addition the entry containing the current
# block is adjusted.
#
proc ::nep::finishBlockMove {w} {
  variable current
  variable cbi
  variable dx
  variable fdof

  #puts "finisBlockMove: cbi=$cbi($w), dx=$dx($w)"

  ## local shortcuts
  set cur $current($w)
  set fd $fdof($w,$cur)

  ## set wx to the current width of the xaxis
  foreach {xmin xmax} [$w.g axis limits x] break
  set wx [expr {$xmax-$xmin}]

  ## compute new xaxis limits
  set xmin [expr {[::tdfio::xCoord $fd $cbi($w)]+$dx($w)}]
  set xmax [expr {$xmin+$wx}]

  $w.g axis configure x -min $xmin -max $xmax
  ::nep::updateDisplay $w

  #setXaxis $w $xvec $old $new
  update idletasks
  event generate "$w.g" <<ReleaseQueue>> -when tail
}
########################################################################
#
# The xaxis was just changed, e.g. by dragging. We set cbi and dx to
# reflect the new view. However if this would pull the data of the
# primary plot out of view, cbi and dx are corrected to reasonable
# values and finishBlockMove is called.
#
proc ::nep::calculateCbiAndDx {w} {
  variable fdof
  variable current
  variable bsof
  variable dx
  variable cbi

  ## local shortcuts
  set cur $current($w)
  
  foreach {xmin xmax} [$w.g axis limits x] break
   
  ## First we make sure that the primary plot is not moved out of
  ## view. If it is, we first correct xmin and xmax again, overriding
  ## whatever the user tried to put upon us. If we don't mess with
  ## xmin and xmax, an updateDisplay suffices to finish our work,
  ## otherwise we have to go through finishBlockMove.
  set finish updateDisplay
  set xlast [::tdfio::xCoord $fdof($w,$cur) last]
  set wx [expr {$xmax-$xmin}]
  if {$xlast<$xmax} {
    set xmax $xlast
    set xmin [expr {$xmax-$wx}]
    set finish finishBlockMove
  }

  ## Now (or anyway) xmin might be too large.
  set x0 [::tdfio::xCoord $fdof($w,$cur) 0]
  if {$xmin<$x0} {
    set xmin $x0
    set xmax [expr {$xmin+$wx}]
    set finish finishBlockMove
  }

  ## find new cbi and its x-coordinate and set dx
  set rs [::tdfio::rSample $fdof($w,$cur) $xmin]
  setCbi $w [expr {$rs-$rs%$bsof($w,$cur)}]
  set xcbi [::tdfio::xCoord $fdof($w,$cur) $cbi($w)]
  set dx($w) [expr $xmin-$xcbi]
  #puts "after calculateCbiAndDx: cbi=$cbi($w), dx=$dx($w)"

  $finish $w
}
########################################################################
#
# Command invoked to initiate either scaling or moving of one of the
# axes. 
# 
proc ::nep::axisCmdStart {w tag cmd x y} {

  ## find the real world coordinates 
  set X [$w.g xaxis invtransform $x]
  set Y [$w.g yaxis invtransform $y]

  ## find pixel coordinates of the axis limits
  foreach {xmin bla} [$w.g axis limits x] break
  foreach {ymin bla} [$w.g axis limits y] break
  set xpmin [$w.g axis transform x $xmin]
  set ypmin [$w.g axis transform y $ymin]
  set padx [lindex [$w.g cget -plotpadx] 0]
  set pady [lindex [$w.g cget -plotpady] 0]
  #puts "$x<->$xpmin-$padx, $y<->$ypmin+$pady"

  ## If we are off the yaxis, only the xaxis is triggered. Note that
  ## pixelcoords of y count from top to bottom (arrrgh!).
  if { $y>=$ypmin+$pady } {
    if { $x<=$xpmin-$padx } return
    #puts "Xaxis only"
    bind $w.g $tag "$cmd {$w} xaxis %x $X"
    bind $w.g <ButtonRelease> \
	[subst {
	  ::nep::calculateCbiAndDx {$w}
	  bind $w.g $tag {}
	  bind $w.g <ButtonRelease> {}
	}]
    return
  } 
  
  ## If we are off the xaxis, only the yaxis is triggered.
  if { $x<=$xpmin-$padx } {
    #puts "Yaxis only"
    bind $w.g $tag "$cmd {$w} yaxis %y $Y"
    bind $w.g <ButtonRelease> \
	[subst {
	  bind $w.g $tag {}
	  bind $w.g <ButtonRelease> {}
	}]
    return
  }

  ## If we are somewhere in the plot, a zoom or unzoom is performed
  if {"$cmd"=="::nep::moveAxis"} {
    #puts zooming
    ::nep::zoomStart $w $X $Y
  } else {
    showPos $w $x $y
  }  
}
########################################################################
#
# Command to move the xaxis such that it tracks the mouse pointer
#
proc ::nep::moveAxis {w axis p Porig} {
  set P [$w.g $axis invtransform $p]
  set delta [expr $P-$Porig]

  set l [$w.g $axis limits]
  set min [lindex $l 0]
  set max [lindex $l 1]
  $w.g $axis configure \
      -min [expr $min-$delta] \
      -max [expr $max-$delta]
}
########################################################################
#
# Command to scale the axis such that it tracks the mouse pointer
# 
proc ::nep::scaleAxis {w axis p Porig} {
  set P [$w.g $axis invtransform $p]

  set l [$w.g $axis limits]
  set min [lindex $l 0]
  set max [lindex $l 1]

  #puts "$P, $Porig"
  if { $P<$min } return

  set s [expr ($P-$min)/($Porig-$min)]
  if {$s<0.01} return

  $w.g $axis configure \
      -max [expr $min+($max-$min)/$s]
}
########################################################################
#
# jump to the block the number of which is found in $w.curblk. Make
# noise, if the entry does not contain a valid block number.
#
proc ::nep::gotoBlock {w} {
  variable bsof
  variable current
  variable fdof

  set bno [$w.curblk get]
  if {[catch {incr bno 0}] || $bno<0} {
    bell 
    return
  }
  
  ## local shortcuts
  set cur $current($w)
  set count [::tdfio::getCount $fdof($w,$cur)]

  set newcbi [expr {$bno*$bsof($w,$cur)}]
  if {$newcbi>=$count} {
    bell
    return
  }
  setCbi $w $newcbi
  finishBlockMove $w
}
########################################################################
#
proc ::nep::showPosSelection {w marker} {
  set text [$w.g marker cget $marker -text]
  selection clear 
  selection handle $w.g [subst {
    return $text
  }]

  set fg [$w.g marker cget $marker -fg]
  $w.g marker configure $marker -fg \
      [$w.g.xsel cget -highlightcolor]
  #puts [$w.g marker configure $marker]

  ## The catch is necessary, because the marker may long have been gone.
  selection own -command [subst {
    catch {$w.g marker configure $marker -fg $fg}
  }] $w.g 
}

########################################################################
#
# Find the closest point of the primary plot and put a marker on it
# denoting its coordinates.
#
proc ::nep::showPos {w x y} {
  variable current
  variable cbi
  variable base
  variable fdof
  variable bsof

  ## local shortcuts
  set cur $current($w)
  set a [$w.g element closest $x $y res -halo 15 $cur]
  if {!$a} return

  #parray res
  set xr $res(x)
  set yr $res(y)

  ## compute the x-value block-relative
  #puts "cbi=$cbi($w)"
  set rs [::tdfio::rSample $fdof($w,$cur) $xr]
  set b [expr {$rs-$rs%$bsof($w,$cur)}]
  set xbr [expr {$xr-[::tdfio::xCoord $fdof($w,$cur) $b]}]

  ## compute anchor so that the label does not pop up with the mouse
  ## in it and immediately disappears
#   foreach {px py} [$w.g transform $xr $yr] break
#   if {$py > $y} {
#     set anchor nw
#     set sign +
#   } else {
#     set sign -
#     set anchor sw
#   }

  set m(0) [$w.g marker create text \
		-name pos.$cur.[uniqueID] \
		-text $xr \
		-coords [list $xr $yr] \
		-anchor se \
		-fill {} \
		-xoffset 3 \
		-yoffset 0 \
		-element $cur \
		-bindtag showPos]
  set m(1) [$w.g marker create text \
		-name pos.$cur.[uniqueID] \
		-text $yr \
		-coords [list $xr $yr] \
		-anchor sw \
		-fill {} \
		-xoffset 0 \
		-yoffset 0 \
		-element $cur]
  set m(2) [$w.g marker create text \
		-name pos.$cur.[uniqueID] \
		-text $xbr \
		-coords [list $xr $yr] \
		-anchor ne \
		-fill {} \
		-xoffset 3 \
		-yoffset 0 \
		-element $cur]
  set m(3) [$w.g marker create text \
		-name pos.$cur.[uniqueID] \
		-text [expr {$b/$bsof($w,$cur)}] \
		-coords [list $xr $yr] \
		-anchor nw \
		-fill {} \
		-xoffset 0 \
		-yoffset 0 \
		-element $cur]
  set m(4) [$w.g marker create bitmap \
		-name pos.$cur.[uniqueID] \
		-bitmap @$base/mark.bm \
		-coords "$xr $yr" \
		-anchor c \
		-element $cur]
  foreach x {0 1 2 3} {
    $w.g marker bind $m($x) <Shift-1> \
	[list ::nep::showPosSelection $w $m($x)]
  }
  foreach x {0 1 2 3 4} {
    $w.g marker bind $m($x) <Shift-2> \
	[list $w.g marker delete $m(0) $m(1) $m(2) $m(3) $m(4)]
  }

}
########################################################################
proc ::nep::deleteAllPos {w} {
  set l [$w.g marker names pos.*]
  eval $w.g marker delete $l
}
########################################################################
#
# set cbi to 0.
#
proc ::nep::Home {w} {
  variable current
  variable cbi
  variable dx
  variable repeat

  ## catch high-frequency repetition of events
  if {!$repeat($w)} return
  set repeat($w) 0

  ## local shortcuts
  set cur $current($w)

  setCbi $w 0
  foreach {xmin xmax} [$w.g axis limits x] break
  set wx [expr {$xmax-$xmin}]
  if {-$dx($w)>=$wx} {
    ## Ooops, we would not see anything of the signal. 
    set dx($w) [expr {-fmod(-$dx($w),$wx)}]
  }
  
  finishBlockMove $w
}
########################################################################
#
# set cbi to the last block of the primary
#
proc ::nep::End {w} {
  variable current
  variable fdof
  variable cbi
  variable dx
  variable bsof
  variable repeat

  ## catch high-frequency repetition of events
  if {!$repeat($w)} return
  set repeat($w) 0

  ## local shortcuts
  set cur $current($w)
  set count [::tdfio::getCount $fdof($w,$cur)]

  set newcbi [expr {($count/$bsof($w,$cur))*$bsof($w,$cur)}]
  if {$newcbi==$count} {
    incr newcbi -$bsof($w,$cur)
  }
  setCbi $w $newcbi
  #puts "End -> cbi=$cbi($w)"
  finishBlockMove $w
} 
########################################################################
#
# Add or substract the the blocksize of the primary to/from cbi
#
# PARAMETER:
# dir -- either "+" or "-" for moving the viewport right or left
#
proc ::nep::RightLeft {w dir} {
  variable cbi
  variable dx
  variable bsof
  variable fdof
  variable current
  variable repeat

  ## local shortcuts
  set cur $current($w)
  set fd $fdof($w,$cur)

  ## The left or right move is performed by moving cbi by one block
  set newcbi [expr "$cbi($w) $dir $bsof($w,$cur)"]
  if {"$dir"=="+"} {
    ## If the new first sample is to far east, we just return.
    if {$newcbi>=[::tdfio::getCount $fd]} return
  } else {
    if {$newcbi<0} return
  }

  ## catch high-frequency repetition of events
  if {!$repeat($w)} return
  set repeat($w) 0
  
  setCbi $w $newcbi
  finishBlockMove $w
}
########################################################################
#
# Change cbi and dx such that the block is shown fully, which contains
# about the middle of the current view. The yrange is not changed.
#
proc ::nep::showFullBlock {w} {
  variable current
  variable fdof
  variable bsof
  variable zstack
  variable dx

  ## local shortcut
  set cur $current($w)
  set bs $bsof($w,$cur)
  set fd $fdof($w,$cur)

  foreach {xmin xmax} [$w.g axis limit x] break
  set ms [::tdfio::rSample $fd [expr {($xmin+$xmax)/2.0}]]
  set newcbi [expr {$ms-$ms%$bs}]
  setCbi $w $newcbi
  #puts $newcbi
  set xmin [::tdfio::xCoord $fd $newcbi]
  set xmax [::tdfio::xCoord $fd [expr {$newcbi+$bs}]]
  #puts [expr $xmax-$xmin]
  set dx($w) 0.0
  $w.g axis configure x -min $xmin -max $xmax
  updateDisplay $w
  set zstack($w) {}
}
########################################################################
#
# Min and max of the yaxis were just changed. Some things have to be
# adapted accordingly.
#
proc ::nep::changedYaxis {w} {
  variable zstack

  set zstack($w) {}
}
########################################################################
#
# Let blt autoscale the current view. Then fix the found values.
#
proc ::nep::autoScaleOnAll {w} {

  $w.g axis configure y -min {} -max {}
  foreach {ymin ymax} [$w.g axis limits y] break

  ## I set the axis limits again because automatically changing axes are 
  ## a good way to shoot you in your foot.
  $w.g axis configure y -min $ymin -max $ymax
  changedYaxis $w
}
########################################################################
# 
# Lets see everything of the primary signal.
#
proc ::nep::fullSignal {w} {
  variable fdof
  variable zstack
  variable current

  set cur $current($w)

  set xmin [::tdfio::xCoord $fdof($w,$cur) 0]
  set xmax [::tdfio::xCoord $fdof($w,$cur) last]

  $w.g xaxis configure -min $xmin -max $xmax
  set zstack($w) {}
  calculateCbiAndDx $w
}
########################################################################
#
# Set the yaxis to minimum and maximum of the current primary block
#
proc ::nep::autoScaleOnPrimary {w} {
  variable current

  set id [string range $current($w) 1 end]
  $w.g axis configure y \
      -min [vector expr min(Y$id)] \
      -max [vector expr max(Y$id)]
  changedYaxis $w
}
########################################################################
#
# Command invoked to initiate a zooming in.
# X and Y are real-world coordinates
proc ::nep::zoomStart {w X Y} {
  global base

  $w.g configure -cursor "@$base/zoomCursor.bm white"
  
  ## the zoom rectangle is painted as a marker, so create the marker
  $w.g marker create polygon \
      -name zoomRect \
      -linewidth 1 \
      -fill {} \
      -outline white \
      -dashes {10 3}

  ## prepare the bindings to move and finish this operation
  bind $w.g <Button1-Motion> "::nep::zoomTrack {$w} zoomRect $X $Y %x %y"
  bind $w.g <ButtonRelease> "::nep::zoomDone {$w} zoomRect $X $Y %x %y"
}
########################################################################
#
# Command to paint a rectangle which follows the mouse pointer
#
proc ::nep::zoomTrack {w rect X0 Y0 x y} {

  set X1 [$w.g xaxis invtransform $x]
  set Y1 [$w.g yaxis invtransform $y]

  $w.g marker configure $rect \
      -coords "$X0 $Y0 $X0 $Y1 $X1 $Y1 $X1 $Y0"
}
########################################################################
#
# Command called when a zoom operation is finished
#
proc ::nep::zoomDone {w rect X0 Y0 x y} {
  variable zstack

  ## reset several things to normal state
  resetCursor $w.g
  bind $w.g <ButtonRelease> {}
  bind $w.g <Button1-Motion> {}
  $w.g marker delete $rect

  ## get the other corner of the rectangle
  set X1 [$w.g xaxis invtransform $x]
  set Y1 [$w.g yaxis invtransform $y]

  ## if the rectangle is too small, don't change a thing.
  if {$X1==$X0 || $Y1==$Y0} return

  ## store the old axis limits on the zoom-stack
  lappend zstack($w) \
      [concat [$w.g axis limits x] [$w.g axis limits y]]

  if {$X0<$X1} {
    $w.g xaxis configure -min $X0 -max $X1
  } else {
    $w.g xaxis configure -min $X1 -max $X0
  }
  if {$Y0<$Y1} {
    $w.g yaxis configure -min $Y0 -max $Y1
  } else {
    $w.g yaxis configure -min $Y1 -max $Y0
  }
  
  ## We have messed with the axis, consequently cbi and dx need an
  ## update 
  calculateCbiAndDx $w
}
########################################################################
#
# Take one element off the zoom stack and configure the plot to show
# exactly that.
#
proc ::nep::unZoom {w} {
  variable zstack

  set l [llength $zstack($w)]
  if {$l==0} {
    bell
    balloonMsg .msg "Nothing to unzoom"
    return
  }

  #puts $zstack($w)
  foreach {xmin xmax ymin ymax} [lindex $zstack($w) end] break
  set zstack($w) [lrange $zstack($w) 0 [expr $l-2]]

  $w.g axis configure x -min $xmin -max $xmax
  #setCurblk $W [currentXvector $W] [lindex $borders 0]
  $w.g axis configure y -min $ymin -max $ymax
  calculateCbiAndDx $w
}
########################################################################
#
# delete the primary plot
#
# Several places must be cleaned up:
# -- the element must be deleted from the graph
# -- the markers associated with the element must be deleted
# -- the menu entry must be deleted from $w.eleMenu.m
# -- the variables associated with the element must be unset
#
# Finally, another element must be selected to be primary.
#
# Parameter which denotes the element in $w.eleMenu.m to which all
# this relates.
#
proc ::nep::deletePlot {w which {force 0}} {
  variable colors

  ## If there is only one plot left, this is forbidden, except force
  ## was set to true
  set lastPlot [$w.eleMenu.m index end]
  if {$lastPlot==0 && !$force} {
    bell
    return
  }
  ## extract the element bound to `which'
  set ele [$w.eleMenu.m entrycget $which -command]

  ## give back the color
  set c [$w.g element cget $ele -color]
  incr colors($c) -1

  ## Delete the vectors and the element from the plot
  set xvec [$w.g element cget $ele -xdata]
  set yvec [$w.g element cget $ele -ydata]
  $w.g element delete $ele
  rename $xvec {}
  rename $yvec {}

  ## Delete associated markers
  set names [$w.g marker names pos.$ele.*]
  eval $w.g marker delete $names

  ## Delete values from variables
  variable bsof
  variable fdof
  unset bsof($w,$ele)
  ::tdfio::close $fdof($w,$ele)
  unset fdof($w,$ele)

  ## Delete it from the menu and activate the next one which follows,
  ## if any.
  set which [$w.eleMenu.m index $which]		;# normalize to
						 # numerical index
  $w.eleMenu.m delete $which
  if {$lastPlot==0} return	;#no more plots available

  if {$which==$lastPlot} {
    ## We just deleted the last one in the row, start over in the front
    set which 0
  }
  elePrimary $w $which
}

########################################################################
#
#
#
proc ::nep::deleteAllPlots {w} {
  ## All plots can be reached via a menu entry in $w.eleMenu.m
  ## We delete them from bottom to the top, because otherwise indexes
  ## would change while we are working on it.
  set i [$w.eleMenu.m index end]
  if {"$i"=="none"} return

  while {$i>=0} {
    deletePlot $w $i 1
    incr i -1
  }
}
########################################################################
#
# addPlot -- add contents of a tdf-file to the graph 
#
proc ::nep::addPlot {w file {raw 0}} {
  variable fdof
  variable bsof

  set id [uniqueID]
  set c [findColor]

  ## attach file
  set fd [::tdfio::open $file $raw]
  
  ## create vectors, currently they are empty
  vector create X$id Y$id
  #puts [namespace which X$id]

  ## create plot element
  $w.g element create g$id \
      -xdata X$id \
      -ydata Y$id \
      -symbol {} \
      -color $c \
      -label [file tail $file]
  $w.g legend configure -hide 0

  ## configure plot specific variables
  set fdof($w,g$id) $fd

  ## Handle blocksize and default xaxis limits
  if {[catch {::tdfio::get $fd Blocksize} bs]} {
    ## no blocksize given
    set bs [::tdfio::getCount $fd]
    if {$bs>=10000} {set bs 5000}
  }
  set bsof($w,g$id) $bs

  ## now for the xaxis
  incr bs -1
  set xmax [::tdfio::xCoord $fd $bs]
  set xmin [::tdfio::xCoord $fd 0]
  $w.g xaxis configure -min $xmin -max $xmax

  eleMenuAdd $w g$id [file tail $file]
 
  ## handle items
  set itmFile [file root $file].itm
  if {[file exist $itmFile]} {
    set itmof($w,g$id) [items::add $w $itmFile]
  }

  ::nep::updateDisplay $w
  autoScaleOnPrimary $w
} 
########################################################################
##
## reload the primary plot
##
## Parameter `which' denotes the entry in $w.eleMenu.m to which all this
## relates. If this is the empty string, the current plot is what we
## want and we have to figure out which entry in the menu it relates
## to.
##
proc ::nep::reloadFile {w {which {}} } {
  variable current
  variable fdof

  set cur $current($w)

  if {""=="$which"} {
    set last [$w.eleMenu.m index end]
    for {set i 0} {$i<=$last} {incr i} {
      if {"$cur"=="[$w.eleMenu.m entrycget $i -command]"} {
	set which $i
	break
      }
    }
    #puts "elemnt: $ele, which: $which"
  } else {
    set cur [$w.eleMenu.m entrycget $which -command]
  }
  set fd $fdof($w,$cur)
  set file [::tdfio::getName $fd]
  set raw  [::tdfio::getRaw $fd]

  deletePlot $w $which 1
  addPlot $w $file $raw
}
########################################################################
#
#
#
proc ::nep::loadFile {w raw} {
  set x [expr [winfo rootx $w.g]+30]
  set y [expr [winfo rooty $w.g]+30]
  set fname [::welDa:: $w.loadFile $x $y ]
  if {""=="$fname"} return

  addPlot $w $fname $raw
}
########################################################################
proc ::nep::loadItems {w} {
  set x [expr [winfo rootx $w.g]+30]
  set y [expr [winfo rooty $w.g]+30]
  set fname [::welDa:: $w.loadItems $x $y *.itm]
  if {""=="$fname"} return

  items::add $w $fname
}
########################################################################
#
#
#
proc ::nep::postScript {w} {
  set x [expr [winfo rootx $w.g]+30]
  set y [expr [winfo rooty $w.g]+30]
  set fname [::welDa:: $w.postScript $x $y *.ps]
  if {""=="$fname"} return

  ## We don't want to see the two little windows in the upper left on
  ## the output.
  $w.g marker configure $w.g.fsel \
      -hide 1

  $w.g postscript output $fname \
      -decorations 0 \
      -landscape 1 \
      -maxpect 1 \
      -height 500 \
      -width 707 \
      -paperheight 29.7c \
      -paperwidth 21.0c

  ## Show the selection windows again.
  $w.g marker configure $w.g.fsel \
      -hide 0
}
########################################################################
#
# Put the legend either inside or outside of the plot
#
proc ::nep::positionLegend {w} {

  switch -exact -- [$w.g legend cget -pos] {
    plotarea {
      $w.g legend configure -pos right -anchor nw
    }
    default {
      $w.g legend configure -pos plotarea -anchor ne
    }
  }
}
########################################################################
#
#
#
proc ::nep::packButtons {w wbuttons} {
  variable buttons

  if {$buttons($w)} {
    pack $wbuttons -side top
  } else {
    pack forget $wbuttons
  }
}
########################################################################
#
# nepExit -- Since nep can be used as a proc, calling exit is no
# solution. Instead we clean up and destroy the window
# 
proc ::nep::Exit {top w} {
  deleteAllPlots $w

  foreach var {modalWait zstack repeat current  \
		   dx cbi legendInside buttons} {
    variable $var
    eval unset $var\($w\)
  }
  destroy $top

  if 1 {
    ## This is debug-output. In an application containing just one
    ## nep-widget, nothing should be printed. If more than one
    ## nep-widget is used and one is destroyed, the arrays of the other
    ## ones must show up.
    foreach x [concat [info vars ::nep::*] [info vars ::nep*]] {
      switch $x {
	::nep::colors { }
	default {
	  catch {parray $x}
	}
      }
    }
  }
} 
########################################################################
#
# setSelection -- let the given window own the selection and highlight
# it. 
#
# This function is intended to be used by the little lables used to
# exchange axis-ranges via the selection.
#
proc ::nep::setSelection {w which} {
  set ws $w.g.${which}sel
  $ws configure -background [$ws cget -highlightbackground]
  $ws configure -foreground [$ws cget -highlightcolor]
  selection handle $ws [subst -nocommands {
    return [$w.g axis limits $which]
  }]
  selection own -command "::nep::looseSelection $ws" $ws
}
proc ::nep::looseSelection {ws} {
  $ws configure -background \
      [lindex [$ws configure -background] 3]
  $ws configure -foreground \
      [lindex [$ws configure -foreground] 3]
}
proc ::nep::setAxisFromSelection {w axis} {
  set g $w.g

  if [catch {selection get} sel] return

  ## is this a well formed list with exactly two elements
  if {[catch {llength $sel} len] || $len!=2} {
    bell
    return
  }
  
  foreach {min max} $sel break
  if {[catch {expr $min}] || [catch {expr $max}] || $max<=$min} {
    bell
    return
  }
  $g axis configure $axis -min $min -max $max
  switch -exact -- $axis {
    x { calculateCbiAndDx $w }
    y { changedYaxis $w}
    default { error "this should not happen" }
  }
}
########################################################################
proc ::nep::resetCursor {w} {
  $w configure -cursor [option get $w [winfo name $w] Cursor]
}
########################################################################
#
#
#
proc ::nep::nep {w geom} {
  variable base
  variable zstack
  variable modalWait

  set zstack($w) {}
  set modalWait($w) 0;                 # to be able to delete it without 
				       # error

  ##
  ## Prepare toplevel window.
  ## If $w is the empty sttring, we assume that nep is supposed to
  ## occupy the top-level window `.', which is supposed to exist
  ## already. If $w is not the empty string, it must be a window name.
  ## That window is either mapped and raised, if it already exists, or
  ## it is created as a toplevel and filled with nep.
  ##
  if {"$w"==""} {
    set top .
  } else {
    set top $w
    if {[winfo exist $w]} {
      wm deiconify $w
      raise $w
      return
    }
    toplevel $top
    if {"$geom"!=""} {
      puts "Setting geom to `$geom'"
      wm geometry $top $geom
    }
  }

  ##
  ## menu bar
  ##
  frame $w.mbar -relief raised -borderwidth 2

  ##
  ## file menu button and its menu
  ##
  menubutton $w.file -text File -menu $w.file.m
  menu $w.file.m -tearoff 0
  $w.file.m add command -label "Load ..." \
      -command "::nep::loadFile \"$w\" 0"
  $w.file.m add command -label "Load raw ..." \
      -command "::nep::loadFile \"$w\" 1"
  $w.file.m add command -label "Load Items ..." \
      -command "::nep::loadItems {$w}"
  $w.file.m add command -label "PostScript ..." \
      -command "::nep::postScript \"$w\""
  $w.file.m add command -label "Quit (q)" \
      -command "::nep::Exit $top \"$w\""
  pack $w.file -in $w.mbar -side left

  ##
  ## Element menu
  ##
  menubutton $w.eleMenu \
      -text Plots \
      -menu $w.eleMenu.m \
      -indicatoron 1 \
      -relief flat 
  pack $w.eleMenu -in $w.mbar -side left
  eleMenuCreate $w

  ##
  ## item manipulation
  ##
  menubutton $w.items -text Items \
      -state disabled \
      -menu $w.items.m
  ::nep::items::createMenu $w $w.items.m
  pack $w.items -in $w.mbar -side left

  ##
  ## miscellaneous functions
  ##
  menubutton $w.misc -text Misc -menu $w.misc.m
  menu $w.misc.m -tearoff 0
  $w.misc.m add checkbutton -label "Legend inside" \
      -variable ::nep::legendInside($w) \
      -onvalue 1 \
      -offvalue 0 \
      -command "::nep::positionLegend \"$w\""
  $w.misc.m add checkbutton -label "Buttons" \
      -variable ::nep::buttons($w) \
      -command "nep::packButtons \"$w\" $w.bframe"
  $w.misc.m add checkbutton -label "balloon help" \
      -variable balloonHelp -onvalue 1
  $w.misc.m add command -label "Delete Positions" \
      -command [list ::nep::deleteAllPos $w]
  if {[llength [info command ::teacher.Create]]} {
    $w.misc.m add command -label Teacher \
	-command "teacher.Create .teacher"
  }
  pack $w.misc -in $w.mbar -side left

  ##
  ## Current-Block entry
  ##
  ##
  label $w.curblkL -text " Block:" -justify right
  entry $w.curblk -width 10
  pack $w.curblk $w.curblkL -in $w.mbar -side right

  ## graph widget and legend
  ## Note that options are set at the beginning of this file with the
  ## option command.
  graph $w.g -highlightthickness 0
  $w.g legend configure 
  $w.misc.m invoke "Legend inside"

  ##
  ## selection set/get buttons for the axes
  ##
  frame $w.g.fsel -borderwidth 1 -background black
  label $w.g.ysel -text y -cursor top_left_arrow
  label $w.g.xsel -text x -cursor top_left_arrow
  pack $w.g.ysel $w.g.xsel -in $w.g.fsel -side left -ipadx 2
  $w.g marker create window \
      -window $w.g.fsel \
      -name $w.g.fsel \
      -xoffset 5 -yoffset 1 \
      -coords {-Inf Inf} \
      -anchor nw
  bind $w.g.xsel <Button1-Motion> [list ::nep::setSelection $w x]
  bind $w.g.xsel <ButtonRelease-2> "::nep::setAxisFromSelection {$w} x"
  selection handle $w.g.xsel "::nep::deliverAxisSelection $w.g xaxis"
  bind $w.g.ysel <Button1-Motion> [list ::nep::setSelection $w y]
  bind $w.g.ysel <ButtonRelease-2> "::nep::setAxisFromSelection {$w} y"
  selection handle $w.g.ysel "::nep::deliverAxisSelection $w.g yaxis"

  
  ##
  ## buttons
  ##
  frame $w.bframe 
  button $w.home -bitmap @$base/home.bm \
      -command "::nep::Home \"$w\""
  button $w.left -bitmap @$base/left.bm \
      -command "::nep::RightLeft \"$w\" -"
  button $w.right -bitmap @$base/right.bm \
      -command "::nep::RightLeft \"$w\" +"
  button $w.end -bitmap @$base/end.bm \
      -command "::nep::End  \"$w\""
  button $w.bscale -bitmap @$base/fullBlock.bm \
      -command "::nep::showFullBlock \"$w\""
  button $w.byscale -bitmap @$base/bscaleOnY.bm \
      -command [list ::nep::autoScaleOnPrimary $w]
  button $w.fullSig -bitmap @$base/fullSignal.bm \
      -command [list ::nep::fullSignal $w]
  button $w.ascale -bitmap @$base/allScaleOnY.bm \
      -command [list ::nep::autoScaleOnAll $w]
  button $w.unZoom -bitmap @$base/unzoom.bm \
      -command "::nep::unZoom \"$w\""

  pack $w.home $w.left $w.right $w.end \
      $w.bscale $w.byscale $w.ascale $w.unZoom $w.fullSig \
      -in $w.bframe \
      -side left
  
  ##
  ## pack it
  ##
  pack $w.mbar -side top -anchor w -fill x
  pack $w.g -side top -fill both -expand 1
  #pack $w.bframe -side top 

  ##
  ## bindings
  ##
  bind $w.g <<ReleaseQueue>> "set ::nep::repeat($w) 1"
  set ::nep::repeat($w) 1

  wm protocol $top WM_DELETE_WINDOW "::nep::Exit $top \"$w\""

  if 1 {
    #    bind $w.g <KeyPress-Shift_L> \
    #	    "$w.g configure -cursor {@$base/zoomCursor.bm white}"
    #bind $w.g <KeyPress-Shift_L> "$w.g configure -cursor pirate"
    #bind $w.g <KeyRelease-Shift_L> "resetCursor $w.g"
    bind $w.g <1> "
      ::nep::axisCmdStart {$w} <B1-Motion> ::nep::moveAxis %x %y;
    "
    bind $w.g <2> "
      ::nep::axisCmdStart {$w} <B2-Motion> ::nep::scaleAxis %x %y
    "    
    bind $w.g <3> [list ::nep::unZoom $w]
    
    ## For the bindings to work with showPos-markers, note that the
    ## modifier-bindings for $w.g must be explicitely cancelled in order 
    ## to prevent invocation of the non-modifier bindings of $w.g.
    bind $w.g <Shift-1> "#relax"
    bind $w.g <Shift-2> "#relax"
    bind $w.g <Shift-3> "#relax"

  }

  bind $w.g <Home> "$w.home invoke"
  bind $w.g <Left> "$w.left invoke"
  bind $w.g <Right> "$w.right invoke"
  bind $w.g <End> "$w.end invoke"
  bind $w.g <F>   "$w.fullSig invoke"
  bind $w.g <a>   "$w.ascale invoke"
  bind $w.g <b>   "$w.bscale invoke"
  bind $w.g <B>   "$w.byscale invoke"
  bind $w.g <u>   "$w.unZoom invoke"
  bind $w.g <Enter> {focus %W}
  bind $w.g <q>   "$w.file.m invoke Quit*"
  bind $w.g <r>   "::nep::reloadFile {$w}"

  bind $w.curblk <Enter> {focus %W}
  bind $w.curblk <Return> [list ::nep::gotoBlock $w]

  ##
  ## add ballon help
  ##
  balloonHelp $w.g \
"Button-1: on axes: move;   on graph: zoom
Button-2: show position;   Shift-1: mark position; Shift-2: delete position
Button-3: unzoom"

  balloonHelp $w.home "to start of primary plot (Home)"
  balloonHelp $w.left "previous block (Left)"
  balloonHelp $w.right "next block (Right)"
  balloonHelp $w.end   "to end of primary plot (End)"
  balloonHelp $w.bscale \
      "change xrange to full block (b)"
  balloonHelp $w.byscale "scale y to min/max of block (B)"
  balloonHelp $w.fullSig "change xrange to show full primary signal (F)"
  balloonHelp $w.ascale  "scale y to min/max of all signals (a)"
  balloonHelp $w.unZoom "unzoom (u)"
  balloonHelp $w.eleMenu  "select primary plot"
  balloonHelp $w.g.xsel \
      "LEFT-motion selects xaxis range\nMIDDLE sets range from selection"
  balloonHelp $w.g.ysel \
      "LEFT-motion selects yaxis range\nMIDDLE sets range from selection"
  balloonHelp $w.curblk \
      "Display and change block number"
  global balloonHelp; set balloonHelp 0
}

namespace eval ::nep {
  namespace export nep
}
