########################################################################
##
## Dialog -- general dialog-creator 
##
## (C) 1997 Harald Kirsch (kir@iitb.fhg.de)
##               ... and the GPL shall apply
##
## This is much more general than tk_dialog, because it allows you to
## put anything you want into the dialog. The main features of this
## dialog are:
## o It performs a grab on the entire tcl-application requiring the
## user to finish the dialog before (s)he can proceed.
## o It creates the widget only once and merely withdraws it from the
## screen after finishing the dialog. This way, all changes the user
## makes to the dialog's subwidgets are kept and the widget can be
## used to store these informations, i.e. no copying into global
## variables is needed.

## w -		name of toplevel-widget to create
## x -          screen-relative x-position 
## y -          screen-relative y-position 
## title -	wm title to put into the widget
## variable -	name of global variable the change of which releases
##              the grab and withdraws the window. The variable will
##              typically be set be an Ok-button.
## createFunc - name of function which creates the GUI within $w. It
##              is called with $w as its first parameter
## args -	additional parameters to be passed to createFunc

##              
##
## (C) 1997 Harald Kirsch (kir@iitb.fhg.de)
## $Revision: 1.3 $, $Date: 1997/06/25 05:31:54 $
##

########################################################################
proc Dialog {w x y title variable createFunc args} {

  if {![winfo exist $w]} {
    toplevel $w 
    wm title $w $title
    wm geometry $w +$x+$y
    eval $createFunc $w $args
  } else {
    wm deiconify $w
  }

  tkwait visibility $w		;#have to wait for the grab to succeed
  grab $w
  tkwait variable $variable
  grab release $w
  wm withdraw $w
}
########################################################################
