#
# My personal favorite file selection dialog.
# I thought it would be obsolete after Tcl7.6, because it has a
# build-in file selection, but soon discovered that I like this one
# sometimes a bit more. 
#
# $Revision: 1.2 $, $Date: 1997/04/04 05:29:05 $
########################################################################
#
# fsDiag.Rescan
#
# scan the given directory with the given pattern and fill
# the given listboxes
# 
proc fsDiag.Rescan {dirL fileL dir pattern} {
  # puts "Rescan: `$dirL' `$fileL' `$dir' `$pattern'"
  $dirL delete 0 end
  $fileL delete 0 end
  set pwd [pwd]
  cd $dir
  set l [lsort [glob .* *]]
  foreach x $l {
    if {[file isdirectory $x]} {
      $dirL insert end $x
    } elseif {[string match $pattern $x]} {
      $fileL insert end $x 
    }
  }
  cd $pwd
}
########################################################################
proc fsDiag.Cancel {w} {
  global fsDiag_Result
  wm withdraw $w
  set fsDiag_Result($w) {}
}
########################################################################
proc fsDiag.Ok {w fileE} {
  global fsDiag_Result

  set res [$fileE get]
  if { [string match $res {}] } {
    bell
    flush stdout
    return
  }
  wm withdraw $w
  set fsDiag_Result($w) [$w.top.currentDir cget -text]/$res
}
########################################################################
proc fsDiag.Use {yPos fileL fileE} {
  # puts "fsDiag.Use($yPos, $fileL, $fileE)"
  set sel [$fileL nearest $yPos]
  $fileL selection clear 0 end
  $fileL selection set $sel
  set txt [$fileL get $sel]
  $fileE delete 0 end
  $fileE insert 0 $txt
}
########################################################################
proc fsDiag.Chdir {w yPos dirL fileL fileE patE} {
  set sel [$dirL get [$dirL nearest $yPos]]
  set oldDir [$w.top.currentDir cget -text]

  if { [string match . $sel] } {
    set sel $oldDir
  } elseif { [string match .. $sel] } {
    set sel [file dirname $oldDir]
  } else {
    if {[string match / $oldDir]} {
      set sel /$sel
    } else {
      set sel $oldDir/$sel
    }
  }
  $dirL delete 0 end
  $fileL delete 0 end
  $fileE delete 0 end
  fsDiag.Rescan $dirL $fileL $sel [$patE get]
  #  scan $w {.%[^.]} topWin
  $w.top.currentDir configure -text $sel
}
########################################################################
#
# fsDiag.Create
#
# Called to create the file selection dialog. Called only once from
# fsDiag. 
####
proc fsDiag.Create {w pat info} {
  global fsDiag_Result

  set fsDiag_Result($w) {}
  toplevel $w -class fsDiag -cursor left_ptr
  wm minsize $w 1 1

  # Some short-names
  set fileL $w.top.file
  set dirL $w.top.dir
  set fileE $w.file
  set patE $w.bot.pat
  set okB  $w.bot.ok
  set cancelB $w.bot.cancel
  set rescanB $w.bot.rescan

  ## A frame for two listboxes
  frame $w.top -border 0

  ## listbox for directories
  listbox $dirL -yscroll "$w.top.dirScroll set" -relief raised -bd 3 \
      -selectmode single

      #-font -schumacher-*-medium-r-*-*-14-*-75-75-*-*-*-*
  scrollbar $w.top.dirScroll -command "$dirL yview"

  ## listbox for non-directories (i.e. files to select)
  listbox $fileL -yscroll "$w.top.fileScroll set" -relief raised -bd 3
      #-font -schumacher-*-medium-r-*-*-14-*-75-75-*-*-*-*
  scrollbar $w.top.fileScroll -command "$fileL yview"

  ## An entry to enter a file by hand
  entry $fileE -relief ridge -bd 4
      #-font -schumacher-*-medium-r-*-*-14-*-75-75-*-*-*-*

  ## A frame for the buttons
  frame $w.bot -border 0

  ## An entry for the file selection pattern
  entry $patE -relief ridge -bd 4 
  #-font -schumacher-*-medium-r-*-*-14-*-75-75-*-*-*-*
  $patE insert 0 $pat

  ## The current Directory
  label $w.top.currentDir -relief ridge -bd 4 -anchor c -text [pwd]

  ## The buttons
  button $okB -text Ok -command "fsDiag.Ok $w $fileE"
  button $cancelB -text Cancel -command "fsDiag.Cancel $w"
  button $rescanB -text Rescan \
      -command "fsDiag.Rescan $dirL $fileL \
                \[$w.top.currentDir cget -text\] \[$patE get\]"

  ## We pack everything from the inside out
  ## Don't know, if the other way round would be better

  ## pack the listboxes into their frame
  pack $w.top.currentDir -fill x
  pack $w.top.dirScroll -side left -fill y
  pack $dirL -side left -fill both -expand yes
  pack $w.top.fileScroll -side left -fill y
  pack $fileL -side left -fill both -expand yes

  ## pack the buttons together with the pattern
  pack $okB $cancelB $rescanB -side left
  pack $patE -side left -expand yes -fill x

  ## pack the top frame, the file entry and the bottom fram
  pack $w.top -side top -expand yes -fill both
  pack $fileE $w.bot -side top -fill x

  ##### bindings #####

  ## the buttons 
  #bind $okB <Button-1> "fsDiag.Ok $w $fileE"
  #bind $cancelB <Button-1> "fsDiag.Cancel $w"
  #bind $rescanB <Button-1> \
  "fsDiag.Rescan $dirL $fileL \[$w.top.currentDir cget -text\] \[$patE get\]"

  ## New listbox bindings. (Copied from listbox.tcl)
  #foreach win "$fileL $dirL" {
  #  bind $win <B1-Motion> {%W select set [%W nearest %y]} 
  #  bind $win <Shift-1> {%W select set [%W nearest %y]}
  #  bind $win <Shift-B1-Motion> {%W select set [%W nearest %y]}
  #}	
  bind $fileL <ButtonRelease-1> "fsDiag.Use %y $fileL $fileE"
  bind $fileL <Double-ButtonRelease-1> "fsDiag.Ok $w $fileE"
    
  #bind $dirL <ButtonPress-1> \
  #    "fsDiag.Chdir $w %y $dirL $fileL $fileE $patE"
  ## Because fileL reacts on Double-1 by using what was clicked on, we
  ## want dirL do this as well. Otherwise a Double-1 would do two
  ## dir-changes, which would be quite confusing
  bind $dirL <Double-ButtonPress-1> \
      "fsDiag.Chdir $w %y $dirL $fileL $fileE $patE"

  ## the entries
  bind $fileE <Return> "fsDiag.Ok $w $fileE"
  bind $patE <Return> \
      "fsDiag.Rescan $dirL $fileL \[$w.top.currentDir cget -text\] \[$patE get\]"
}
########################################################################
#
# fsDiag -- my favorite file selection dialog
# 
# PARAMETERS:
# wName -- name to use for the whole thing
# screenX -- x-coordinate on screen where dialog is posted.
# screenY -- y-coordinate on screen where dialog is posted.
# pat -- pattern to use when selecting files
# info -- allows additional information about the action taking place
#
# When first called, the dialog is created for the current directory,
# at the given position and the list box for files is filled using the
# given pattern.
#
# However, the state of the dialog box is kept between calls, so on
# subsequent calls the parameters screenX, screenY and pat are
# ignored. 
#####
proc fsDiag {wName screenX screenY {pat *} {info "select file"}} {
  global fsDiag_Result
  # Some short-names
  set fileL $wName.top.file
  set dirL $wName.top.dir
  set fileE $wName.file
  set patE $wName.bot.pat
  set okB  $wName.bot.ok
  set cancelB $wName.bot.cancel
  set rescanB $wName.bot.rescan


  if {![winfo exist $wName]} {
    fsDiag.Create $wName $pat $info
    wm geometry $wName +$screenX+$screenY
  } else {
    wm deiconify $wName
  }

  ## fill the listboxes
  fsDiag.Rescan $dirL $fileL \
      [$wName.top.currentDir cget -text] [$patE get]

  grab $wName
  tkwait variable fsDiag_Result($wName)
  grab release $wName
  return $fsDiag_Result($wName)
}  

########################################################################
if 0 {
  button .b -text "call file selection 1" -command {
    puts "from 1: [fsDiag .x1 500 100 *~ "save as ..."]"
  }
  button .c -text "call file selection 2" -command {
    puts "from 2: [fsDiag .x2 500 100 *.tcl]"
  }
  pack .b .c -side top -padx 10 -pady 10
}

