#!/t/BIN/wish -f
#
# My personal favorite file selection dialog which allows to select a
# whole list of files.
#
#
## global variables:
# multFsDiag -- array containing quite a bit of info, in particular
# multFsDiag(<widget>,res) : result of operation
#
#
# $Revision: 1.1 $, $Date: 1997/05/08 18:07:55 $
########################################################################
#
# multFsDiag.Rescan
#
# scan the given directory with the given pattern and fill
# the given listboxes
# 
proc multFsDiag.Rescan {dirL fileL dir pattern} {
  # puts "Rescan: `$dirL' `$fileL' `$dir' `$pattern'"
  $dirL delete 0 end
  $fileL delete 0 end
  set pwd [pwd]
  cd $dir
  set l [lsort [glob .* *]]
  foreach x $l {
    if {[file isdirectory $x]} {
      $dirL insert end $x
    } elseif {[string match $pattern $x]} {
      $fileL insert end $x 
    }
  }
  cd $pwd
}
########################################################################
proc multFsDiag.Cancel {w} {
  global multFsDiag
  wm withdraw $w
  set multFsDiag($w,res) {}
}
########################################################################
proc multFsDiag.Ok {w fileL} {
  global multFsDiag

  set l [$fileL curselection]
  if ![llength $l] {
    bell 
    return
  }
  set dir [$w.curDir cget -text]
  foreach i $l {
    lappend res $dir/[$fileL get $i]
  }
  wm withdraw $w
  set multFsDiag($w,res) $res
}
########################################################################
proc multFsDiag.Use {yPos fileL fileE} {
  # puts "multFsDiag.Use($yPos, $fileL, $fileE)"
  set sel [$fileL nearest $yPos]
  $fileL selection clear 0 end
  $fileL selection set $sel
  set txt [$fileL get $sel]
  $fileE delete 0 end
  $fileE insert 0 $txt
}
########################################################################
proc multFsDiag.Chdir {w yPos} {
  set dirL $w.top.dir
  set fileL $w.top.file

  set sel [$dirL get [$dirL nearest $yPos]]
  set oldDir [$w.curDir cget -text]

  if { [string match . $sel] } {
    set sel $oldDir
  } elseif { [string match .. $sel] } {
    set sel [file dirname $oldDir]
  } else {
    if {[string match / $oldDir]} {
      set sel /$sel
    } else {
      set sel $oldDir/$sel
    }
  }
  $dirL delete 0 end
  $fileL delete 0 end
  multFsDiag.Rescan $dirL $fileL $sel [$w.pat get]
  #  scan $w {.%[^.]} topWin
  $w.curDir configure -text $sel
}
########################################################################
#
# multFsDiag.Create
#
# Called to create the file selection dialog. Called only once from
# multFsDiag. 
####
proc multFsDiag.Create {w pat info} {
  global multFsDiag

  set multFsDiag($w) {}
  toplevel $w -class MultFsDiag -cursor left_ptr
  wm minsize $w 1 1
  wm title $w $info

  # Some short-names
  set fileL $w.top.file
  set dirL $w.top.dir
  set okB  $w.bot.ok
  set cancelB $w.bot.cancel
  set rescanB $w.bot.rescan
  set selallB $w.bot.selall

  ## label for the current dir
  label $w.curDir -padx 4 -anchor w -text [pwd]



  ## The two listboxes and their scrollbars
  frame $w.top -border 0

  listbox $dirL -yscroll "$w.top.dirScroll set"
  scrollbar $w.top.dirScroll -command "$dirL yview" -width 12

  listbox $fileL \
      -selectmode extended \
      -yscroll "$w.top.fileScroll set"
  scrollbar $w.top.fileScroll -command "$fileL yview" -width 12

  pack $w.top.dirScroll -side left -fill y
  pack $dirL -side left -fill both -expand yes
  pack $w.top.fileScroll -side left -fill y
  pack $fileL -side left -fill both -expand yes



  ## A label and an entry for the file selection pattern
  frame $w.patF -borderwidth 3 -relief ridge
  label $w.patL -text Filter: -anchor e -padx 4
  entry $w.pat  -relief flat \
      -font -dec-*-*-*-*-*-*-*-*-*-*-*-iso8859-*
  $w.pat insert 0 $pat
  pack $w.patL -in $w.patF -side left
  pack $w.pat -in $w.patF -side left -expand 1 -fill x



  ## The buttons
  frame $w.bot -border 0

  button $okB -text Ok -command "multFsDiag.Ok $w $w.top.file"
  button $cancelB -text Cancel -command "multFsDiag.Cancel $w"
  button $rescanB -text Rescan \
      -command \
      "multFsDiag.Rescan $dirL $fileL \[$w.curDir cget -text] \[$w.pat get\]"
  button $selallB -text {Select All} \
      -command "$w.top.file selection set 0 end"

  pack $okB $cancelB $rescanB $selallB -side left


  ## pack everything 
  pack $w.curDir -side top -fill x
  pack $w.top -side top -expand yes -fill both
  pack $w.patF -side top -fill x
  pack $w.bot -side top -anchor c

  ##### bindings #####
  bind $w.pat <Return> "$rescanB invoke"
  bind $dirL <Double-1> "multFsDiag.Chdir $w %y"
}
########################################################################
#
# multFsDiag -- my favorite file selection dialog for multiple files
# 
# PARAMETERS:
# wName -- widget name to use for the whole thing
# screenX -- x-coordinate on screen where dialog is posted.
# screenY -- y-coordinate on screen where dialog is posted.
# pat -- pattern to use when selecting files
# info -- allows additional information about the action taking place
#
# When first called, the dialog is created for the current directory,
# at the given position and the list box for files is filled using the
# given pattern.
#
# However, the state of the dialog box is kept between calls, so on
# subsequent calls the parameters screenX, screenY and pat are
# ignored. 
#####
proc multFsDiag {wName screenX screenY {pat *} {info "select files"}} {
  global multFsDiag
  # Some short-names
  set fileL $wName.top.file
  set dirL $wName.top.dir
  set fileE $wName.file
  set patE $wName.pat
  set okB  $wName.bot.ok
  set cancelB $wName.bot.cancel
  set rescanB $wName.bot.rescan


  if {![winfo exist $wName]} {
    multFsDiag.Create $wName $pat $info
    wm geometry $wName +$screenX+$screenY
  } else {
    wm deiconify $wName
  }

  ## fill the listboxes
  multFsDiag.Rescan $dirL $fileL [$wName.curDir cget -text] [$patE get]

  grab $wName
  tkwait variable multFsDiag($wName,res)
  grab release $wName
  return $multFsDiag($wName,res)
}  

########################################################################
if 0 {
  button .b -text "call file selection 1" -command {
    puts "from 1: `[multFsDiag .x1 500 100 *~ "save as ..."]'"
  }
  button .c -text "call file selection 2" -command {
    puts "from 2: `[multFsDiag .x2 500 100 *.tcl]'"
  }
  pack .b .c -side top -padx 10 -pady 10
}

