########################################################################
##
## Read tdf-header from an open file and enter the key-value pairs
## into the array named $_TDFHEADER. There is no possibility for a
## syntax error, because everthing in a header-line is optional, in
## particular, empty lines are ignored.
##
## PARAMETER:
## in --
## file to read from
##
## _header
## name of the array to put key-value pairs into.
##
## ERRORS:
## -- no binary marker found
## 
## $Revision: 1.1 $, $Date: 1998/06/24 05:17:04 $
########################################################################

proc tdfReadHeader {in _header} {
  upvar $_header header
 
  set res {}
  set ch [read $in 1]
  while {![eof $in] && "$ch"!="\025"} {
    append res $ch
    set ch [read $in 1]
  }
  if {$ch!="\025"} {
    return -code error -errorinfo "no binary marker found"
  }
  set key {}
  set value {}
  foreach line [split $res "\n"] {
    if {"[string trim $line]"==""} continue
    regexp "^ *(\[^ \t:\]*) *:? *(.*)" $line d key value
    #puts "`$key'[string length $key]: `$value'[string length $value]"
    set header($key) $value
  }
}
########################################################################
## This is for tests only. It is executed, if you run this file by
## itself and don't source it.
##
if {"[file tail $argv0]"=="tdfReadHeader.tcl"} {
  set in [open [lindex $argv 0] r]
  tdfReadHeader $in HEAD
  close $in
  parray HEAD
}
