#
# And another file selection dialogue.
#
# I thought it would be obsolete after Tcl7.6, because it has a
# build-in file selection, but soon discovered that I like this one
# sometimes a bit more. 
#
# ----- EXAMPLE AT END OF THIS FILE -----
#
# $Revision: 1.3 $, $Date: 1999/05/25 08:13:01 $
########################################################################

## No need to source this twice.
if {![catch {info body ::welDa::}]} {
  return
}

## all the procs in here are declared within namespace ::welDa
namespace eval ::welDa {

  if {[info exist env(LANG)]} {
    set lang [string tolower $env(LANG)]
  } else {
    set lang en
  }
  switch -glob -- $lang {
    de* {
      option add *WelDa.patlist.text Dateimuster widgetDefault
      option add *WelDa.ok.text Ok widgetDefault
      option add *WelDa.cancel.text Abbruch widgetDefault
    }
    en* {
      option add *WelDa.patlist.text Pattern widgetDefault
      option add *WelDa.ok.text Ok widgetDefault
      option add *WelDa.cancel.text Cancel widgetDefault
    }
  }

variable upButtonBitmap  {
  #define updir_width 28
  #define updir_height 16
  static char updir_bits[] = {
    0x00, 0x00, 0x00, 0x00, 0x80, 0x1f, 0x00, 0x00, 0x40, 0x20, 0x00, 0x00,
    0x20, 0x40, 0x00, 0x00, 0xf0, 0xff, 0xff, 0x01, 0x10, 0x00, 0x00, 0x01,
    0x10, 0x02, 0x00, 0x01, 0x10, 0x07, 0x00, 0x01, 0x90, 0x0f, 0x00, 0x01,
    0x10, 0x02, 0x00, 0x01, 0x10, 0x02, 0x00, 0x01, 0x10, 0x02, 0x00, 0x01,
    0x10, 0xfe, 0x07, 0x01, 0x10, 0x00, 0x00, 0x01, 0x10, 0x00, 0x00, 0x01,
    0xf0, 0xff, 0xff, 0x01};}
}


########################################################################
#
# welDa -- my new favorite file selection dialog
# 
# PARAMETERS:
# w -- widget-name to use for the whole thing
# screenX -- x-coordinate on screen where dialog is posted.
# screenY -- y-coordinate on screen where dialog is posted.
# pat -- pattern to use when selecting files
# info -- allows additional information about the action taking
#         place. It is put into the title-bar of the dialog.
#
# When first called, the dialog is created for the current directory,
# at the given position and the list box for files is filled using the
# given pattern.
#
# However, the state of the dialog box is kept between calls, so on
# subsequent calls the parameters screenX, screenY and pat are
# ignored. 
#####
proc ::welDa:: {w screenX screenY {pat {{* "all files"}}} {info "select file"}} {
  variable Result

  if {![winfo exist $w]} {
    Create $w $pat $info
    wm geometry $w +$screenX+$screenY
  } else {
    set d [$w.pwd get]
    while {![file isdir $d]} {
      set d [file dir $d]
    }
    $w.pwd delete 0 end
    $w.pwd insert end $d
    Expand $w
    wm deiconify $w
  }

  ## fill the listboxes
  Expand $w 1

  grab $w
  vwait ::welDa::Result($w)
  grab release $w
  return $Result($w)
}  
########################################################################
proc ::welDa::Expand {w {recursive 0}} {
  set usertext [string trim [$w.pwd get]]
  if {""=="$usertext"} {
    set usertext [pwd]
  }
  set pat [string trim [$w.pat get]]

  ## initialize some variables
  ## dir -- the dir where glob is called
  ## prefix -- within that dir the required prefix of the file names
  if [string match */ $usertext] {
    set dir [file dir ${usertext}a]
    set prefix {}
  } else {
    set dir [file dir $usertext]
    set prefix [file tail $usertext]
  }

  ## Nevertheless we glob everything and then sort out the wrong
  set l [lsort [glob -nocomplain $dir/.* $dir/*]]

  $w.dirs delete 0 end
  $w.files delete 0 end

  set which {}		;# tail of those that match
  set haveFiles 0	;# set to 1, if there is at least one file
  foreach x $l {
    set t [file tail $x]
    if {"."=="$t" || ".."=="$t"} continue

    if [file isdir $x] {
      if {[string match $prefix* $t]} {
	$w.dirs insert end $t
	lappend which $t
      }
    } elseif {[string match $prefix* $t] && [string match $pat $t]} {
      $w.files insert end $t
      lappend which $t
      set haveFiles 1
    }
  }

  ## prepare $w.pwd
  $w.pwd delete 0 end
  $w.pwd insert end "$dir"
  if {![string match */ $dir]} {$w.pwd insert end /}

  if {$recursive} {
    $w.pwd xview end
    return
  }
  if {[llength $which]==1} {
    $w.pwd insert end [lindex $which 0]
    $w.pwd xview end
    if [file isdir [$w.pwd get]] {
      $w.pwd insert end /
      #puts "recursing for [$w.pwd get]"
      Expand $w 1
    }
    return
  }

  ## find the longest common prefix within all entries of `which'
  set i [string length $prefix]
  set tmp [lindex $which 0]
  set ok 1
  while {""!="[set char [string index $tmp $i]]"} {
    foreach name $which {
      if {"$char"!="[string index $name $i]"} {
	set ok 0
	break
      }
    }
    if {!$ok} break
    append prefix $char
    incr i
  }
  
  $w.pwd insert end $prefix
  $w.pwd xview end
}
########################################################################
proc ::welDa::Updir {w} {
  set dir [$w.pwd get]
  while {![file isdir $dir]} {
    set dir [file dir $dir]
  }
  set tail [file tail $dir]
  set dir [file dir $dir]
  $w.pwd delete 0 end
  $w.pwd insert end $dir
  Expand $w
  
  set len [$w.dirs index end]
  $w.dirs selection clear 0 end
  # puts "looking for $tail in $dir"
  for {set i 0} {$i<=$len} {incr i} {
    if {"$tail"=="[$w.dirs get $i]"} {
      $w.dirs see $i
      $w.dirs selection set $i
      break
    }
  }
  
}
########################################################################
proc ::welDa::Cancel {w} {
  variable Result
  wm withdraw $w
  set Result($w) {}
}
########################################################################
proc ::welDa::Ok {w} {
  variable Result

  #puts "namespace [namespace current] has procs [info procs]"
  #puts "it has vars [info vars]"
  set res [$w.pwd get]
  if { [string match $res {}] } {
    bell
    return
  }
  wm withdraw $w
  set Result($w) $res
}
########################################################################
proc ::welDa::Use {w yPos lbox} {
  #puts "Use($w,$yPos,$lbox)"
  set sel [$lbox get [$lbox nearest $yPos]]
  if [string match *dirs $lbox] {
    append sel /
  }
  set usertext [string trim [$w.pwd get]]

  if [string match */ $usertext] {
    $w.pwd insert end $sel
  } else {
    $w.pwd delete 0 end
    $w.pwd insert end [file dir $usertext]/$sel
  }
  $w.pwd xview end
}
########################################################################
#
# Create
#
# Called to create the file selection dialog. Called only once from
# ::welda:: 
####
proc ::welDa::Create {w pat info} {
  variable Result
  variable upButtonBitmap

  set Result($w) {}
  toplevel $w -class WelDa -cursor left_ptr
  wm minsize $w 1 1
  wm title $w $info

  ## A frame for two listboxes
  frame $w.top -border 0

  ## listbox for directories and its scrollbar
  listbox $w.dirs -yscroll "$w.top.dirScroll set" \
      -selectmode browse \
      -relief sunken 
  scrollbar $w.top.dirScroll -command "$w.dirs yview"

  ## listbox for non-directories (i.e. files to select) and scrollbar
  listbox $w.files -yscroll "$w.top.fileScroll set" \
      -selectmode browse \
      -relief sunken
  scrollbar $w.top.fileScroll -command "$w.files yview"

  ## The middle frame
  frame $w.mid -border 0

  ## The updir button
  button $w.back -command "::welDa::Updir $w" 
  set img [image create bitmap -data $upButtonBitmap \
	       -foreground [$w.back cget -foreground]]
  $w.back configure -image $img

  ## The current Directory. Its bindings are heavily modified. The
  ## main purpose is to change the meaning of the space character
  ## while maintaining all other default bindings.
  entry $w.pwd -relief sunken -bd 2
  $w.pwd insert end "[pwd]/"
  focus $w.pwd
  bind ExpandEntry <space> "::welDa::Expand $w"
  foreach event [bind Entry] {
    bind ExpandEntry $event [bind Entry $event]
  }
  set l [bindtags $w.pwd]
  regsub Entry $l ExpandEntry l
  bindtags $w.pwd $l

  ## The bottom frame
  frame $w.bot -border 0

  ## A label and an entry for the file selection pattern
  set m $w.patlist.m
  menubutton $w.patlist -indicatoron 1 -menu $m -relief raised
  menu $m -tearoff 0
  foreach x $pat {
    foreach {p txt} $x {}
    $m add command -label "$txt: $p" \
	-command [concat $w.pat delete 0 end \; $w.pat insert end $p \; \
		     ::welDa::Expand $w]
  }
  entry $w.pat -relief sunken -bd 2
  $w.pat insert 0 [lindex [lindex $pat 0] 0]

  ## The buttons
  button $w.ok -command "::welDa::Ok $w"
  button $w.cancel -command "::welDa::Cancel $w"

  ## We pack everything from the inside out
  ## Don't know, if the other way round would be better

  ## pack the listboxes into their frame
  pack $w.dirs -in $w.top -side left -fill both -expand yes
  pack $w.top.dirScroll -side left -fill y
  pack $w.files -in $w.top -side left -fill both -expand yes
  pack $w.top.fileScroll -side left -fill y

  ## pack the middle frame
  #pack $w.back -in $w.mid -side left
  #pack $w.pwd -in $w.mid -side left -expand yes -fill x
  grid $w.patlist -in $w.mid -row 0 -column 0 -sticky e
  grid $w.pat     -in $w.mid -row 0 -column 1 -sticky e -pady 5
  grid $w.pwd     -in $w.mid -row 1 -column 0 \
      -columnspan 2 \
      -sticky ew -pady 5
  grid $w.back    -in $w.mid -row 1 -column 2 -sticky w
  grid columnconfigure $w.mid 0 -weight 1

  ## pack the buttons together with the pattern
  #pack $w.patl -in $w.bot -side left
  #pack $w.patlist -in $w.bot -side left
  #pack $w.pat -in $w.bot -side left -expand yes -fill x
  pack $w.ok $w.cancel -in $w.bot -side left

  ## pack the top frame, the file entry and the bottom frame
  pack $w.top -side top -expand yes -fill both
  pack $w.mid -side top -fill x -expand yes
  pack $w.bot -side top -expand yes -fill none

  ##### bindings #####

  ## the buttons 
  bind $w <Key-Escape> "::welDa::Cancel $w"

  bind $w.files <ButtonRelease-1> "::welDa::Use $w %y $w.files"
  bind $w.files <Double-ButtonRelease-1> "::welDa::Ok $w"

  bind $w.dirs <ButtonRelease-1> \
      "::welDa::Use $w %y $w.dirs; ::welDa::Expand $w"

  bind $w.pwd <Return> "::welDa::Ok $w"

  bind $w.pat <Return> "::welDa::Expand $w"
}
########################################################################



if {"$argv"=="welDa"} {
  wm withdraw .
#  set l {{*.tcl {tcl-Dateien}} {* {alle Dateien}}}
  set l {{*.tcl {tcl-Dateien}}}
  set res [::welDa:: .x 100 100 $l "select file"]
  if {""=="$res"} {
    puts "You canceled the dialogue."
  } else {
    puts "You selected `$res'"
  }
  exit 0
}
